/*
Copyright (C) Arturo Torres Sánchez 2013

This file is part of Trino.

Trino is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Trino is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Trino.  If not, see <http://www.gnu.org/licenses/>.
*/

namespace Trino {
    public class TweetRegex {
      private static string PUNCT = "\\x{0021}-\\x{002F}\\x{003A}-\\x{0040}\\x{005B}-\\x{0060}\\x{007B}-\\x{007E}";
      private static string SPACES = " \\t\\n\\x{0B}\\f\\r";
      private static string CTRL = "\\x{00}-\\x{1F}\\x{7F}";
      private static string INVALID = "\\x{FFFE}\\x{FEFF}\\x{FFFF}\\x{202A}-\\x{202E}";
      private static string UNICODE_SPACES = 
        "\\x{0009}-\\x{000d}" +   //  # White_Space # Cc   [5] <control-0009>..<control-000D>
        "\\x{0020}" +             // White_Space # Zs       SPACE
        "\\x{0085}" +             // White_Space # Cc       <control-0085>
        "\\x{00a0}" +             // White_Space # Zs       NO-BREAK SPACE
        "\\x{1680}" +             // White_Space # Zs       OGHAM SPACE MARK
        "\\x{180E}" +             // White_Space # Zs       MONGOLIAN VOWEL SEPARATOR
        "\\x{2000}-\\x{200a}" +   // # White_Space # Zs  [11] EN QUAD..HAIR SPACE
        "\\x{2028}" +             // White_Space # Zl       LINE SEPARATOR
        "\\x{2029}" +             // White_Space # Zp       PARAGRAPH SEPARATOR
        "\\x{202F}" +             // White_Space # Zs       NARROW NO-BREAK SPACE
        "\\x{205F}" +             // White_Space # Zs       MEDIUM MATHEMATICAL SPACE
        "\\x{3000}";              // White_Space # Zs       IDEOGRAPHIC SPACE

      private static string LATIN_ACCENTS_CHARS = "\\x{00c0}-\\x{00d6}\\x{00d8}-\\x{00f6}\\x{00f8}-\\x{00ff}" + // Latin-1
                                                  "\\x{0100}-\\x{024f}" + // Latin Extended A and B
                                                  "\\x{0253}\\x{0254}\\x{0256}\\x{0257}\\x{0259}\\x{025b}\\x{0263}\\x{0268}\\x{026f}\\x{0272}\\x{0289}\\x{028b}" + // IPA Extensions
                                                  "\\x{02bb}" + // Hawaiian
                                                  "\\x{0300}-\\x{036f}" + // Combining diacritics
                                                  "\\x{1e00}-\\x{1eff}"; // Latin Extended Additional (mostly for Vietnamese)
      private static string HASHTAG_ALPHA_CHARS = "a-z" + LATIN_ACCENTS_CHARS +
                                                       "\\x{0400}-\\x{04ff}\\x{0500}-\\x{0527}" +  // Cyrillic
                                                       "\\x{2de0}-\\x{2dff}\\x{a640}-\\x{a69f}" +  // Cyrillic Extended A/B
                                                       "\\x{0591}-\\x{05bf}\\x{05c1}-\\x{05c2}\\x{05c4}-\\x{05c5}\\x{05c7}" +
                                                       "\\x{05d0}-\\x{05ea}\\x{05f0}-\\x{05f4}" + // Hebrew
                                                       "\\x{fb12}-\\x{fb28}\\x{fb2a}-\\x{fb36}\\x{fb38}-\\x{fb3c}\\x{fb3e}\\x{fb40}-\\x{fb41}" +
                                                       "\\x{fb43}-\\x{fb44}\\x{fb46}-\\x{fb4f}" + // Hebrew Pres. Forms
                                                       "\\x{0610}-\\x{061a}\\x{0620}-\\x{065f}\\x{066e}-\\x{06d3}\\x{06d5}-\\x{06dc}" +
                                                       "\\x{06de}-\\x{06e8}\\x{06ea}-\\x{06ef}\\x{06fa}-\\x{06fc}\\x{06ff}" + // Arabic
                                                       "\\x{0750}-\\x{077f}\\x{08a0}\\x{08a2}-\\x{08ac}\\x{08e4}-\\x{08fe}" + // Arabic Supplement and Extended A
                                                       "\\x{fb50}-\\x{fbb1}\\x{fbd3}-\\x{fd3d}\\x{fd50}-\\x{fd8f}\\x{fd92}-\\x{fdc7}\\x{fdf0}-\\x{fdfb}" + // Pres. Forms A
                                                       "\\x{fe70}-\\x{fe74}\\x{fe76}-\\x{fefc}" + // Pres. Forms B
                                                       "\\x{200c}" +                        // Zero-Width Non-Joiner
                                                       "\\x{0e01}-\\x{0e3a}\\x{0e40}-\\x{0e4e}" + // Thai
                                                       "\\x{1100}-\\x{11ff}\\x{3130}-\\x{3185}\\x{A960}-\\x{A97F}\\x{AC00}-\\x{D7AF}\\x{D7B0}-\\x{D7FF}" + // Hangul (Korean)
                                                       "\\x{ffa1}-\\x{ffdc}" +                 // half width Hangul (Korean)
                                                       "\\x{30A1}-\\x{30FA}" + //Katakana (full-width)
                                                       "\\x{ff66}-\\x{ff9f}" + //Katakana (half-width)
                                                       "\\x{ff10}-\\x{ff19}\\x{ff21}-\\x{ff3a}\\x{ff41}-\\x{ff5a}" + //Latin (full-width)
                                                       "\\x{3041}-\\x{3096}\\x{3099}-\\x{309e}" + //Hiragana
                                                       "\\x{3400}-\\x{4DBF}" + //Kanji (CJK Extension A)
                                                       "\\x{4E00}-\\x{9FFF}" + //Kanji (Unified)
                                                       "\\x{20000}-\\x{2A6DF}" + //Kanji (CJK Extension B)
                                                       "\\x{2A700}-\\x{2B73F}" + //Kanji (CJK Extension C)
                                                       "\\x{2B740}-\\x{2B81F}" + //Kanji (CJK Extension D)
                                                       "\\x{2F800}-\\x{2FA1F}\\x{3003}\\x{3005}\\x{303B}"; //Kanji (CJK supplement)
      private static string HASHTAG_ALPHA_NUMERIC_CHARS = "0-9\\x{ff10}-\\x{ff19}_" + HASHTAG_ALPHA_CHARS;
      private static string HASHTAG_ALPHA = "[" + HASHTAG_ALPHA_CHARS +"]";
      private static string HASHTAG_ALPHA_NUMERIC = "[" + HASHTAG_ALPHA_NUMERIC_CHARS +"]";

      /* URL related hash regex collection */
      private static string URL_VALID_PRECEEDING_CHARS = "(?:[^A-Z0-9@＠$#＃_/\\-\\." + INVALID + "]|^)";

      //private static string URL_VALID_CHARS = "[" + ALNUM + LATIN_ACCENTS_CHARS + "]";
      private static string URL_VALID_CHARS = "[^" + PUNCT + SPACES + CTRL + INVALID + UNICODE_SPACES + "]";

      private static string URL_VALID_SUBDOMAIN = "(?:(?:" + URL_VALID_CHARS + "(?:[_-]|" + URL_VALID_CHARS + ")*)?" + URL_VALID_CHARS + "\\.)";
      private static string URL_VALID_DOMAIN_NAME = "(?:(?:" + URL_VALID_CHARS + "(?:[-]|" + URL_VALID_CHARS + ")*)?" + URL_VALID_CHARS + "\\.)";
      /* Any non-space, non-punctuation characters. \p{Z} = any kind of whitespace or invisible separator. */
      //private static string URL_VALID_UNICODE_CHARS = "[.[^"+ PUNCT +"\\s\\p{Z}\\x{2000}-\\x{206F}]]";

      private static string URL_VALID_GTLD =
          "(?:(?:aero|asia|biz|cat|com|coop|edu|gov|info|int|jobs|mil|mobi|museum|name|net|org|pro|tel|travel|xxx)(?=[^0-9a-z]|$))";
      private static string URL_VALID_CCTLD =
          "(?:(?:ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|" +
          "bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|" +
          "er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|" +
          "hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|" +
          "lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|" +
          "nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|sk|" +
          "sl|sm|sn|so|sr|ss|st|su|sv|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|" +
          "va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|za|zm|zw)(?=[^0-9a-z]|$))";
      private static string URL_PUNYCODE = "(?:xn--[0-9a-z]+)";

      /*private static string URL_VALID_DOMAIN =
        "(?:" +                                                   // subdomains + domain + TLD
            URL_VALID_SUBDOMAIN + "+" + URL_VALID_DOMAIN_NAME +   // e.g. www.twitter.com, foo.co.jp, bar.co.uk
            "(?:" + URL_VALID_GTLD + "|" + URL_VALID_CCTLD + "|" + URL_PUNYCODE + ")" +
          ")" +
        "|(?:" +                                                  // domain + gTLD
          URL_VALID_DOMAIN_NAME +                                 // e.g. twitter.com
          "(?:" + URL_VALID_GTLD + "|" + URL_PUNYCODE + ")" +
        ")" +
        "|(?:" + "(?<=http://|https://)" +
          "(?:" +
            "(?:" + URL_VALID_DOMAIN_NAME + URL_VALID_CCTLD + ")" +  // protocol + domain + ccTLD
            "|(?:" +
              URL_VALID_CHARS + "+\\." +                     // protocol + Unicode domain + TLD
              "(?:" + URL_VALID_GTLD + "|" + URL_VALID_CCTLD + ")" +
            ")" +
          ")" +
        ")" +
        "|(?:" +                                                  // domain + ccTLD + '/'
          URL_VALID_DOMAIN_NAME + URL_VALID_CCTLD + "(?=/)" +     // e.g. t.co/
        ")";*/
      private static string URL_VALID_DOMAIN =
        "(?:" + 
          @"$URL_VALID_SUBDOMAIN+$URL_VALID_DOMAIN_NAME" + 
          @"(?:$URL_VALID_GTLD|$URL_VALID_CCTLD|$URL_PUNYCODE)" + 
        ")|(?:" + 
          @"$URL_VALID_DOMAIN_NAME" + 
          @"(?:$URL_VALID_GTLD|$URL_PUNYCODE)" + 
        ")|(?:" + 
          @"$URL_VALID_DOMAIN_NAME$URL_VALID_CCTLD(?=/)" + 
        ")|(?:" + 
          "(?<=http://|https://)" +
          @"$URL_VALID_SUBDOMAIN*$URL_VALID_DOMAIN_NAME" + 
          @"(?:$URL_VALID_GTLD|$URL_VALID_CCTLD|$URL_PUNYCODE)" + 
        ")";

      private static string URL_VALID_PORT_NUMBER = "[0-9]++";

      private static string URL_VALID_GENERAL_PATH_CHARS = "[a-z0-9!\\*';:=\\+,\\.\\$/%#\\[\\]\\-_~\\|&" + LATIN_ACCENTS_CHARS + "]";
      /** Allow URL paths to contain balanced parens
       *  1. Used in Wikipedia URLs like /Primer_(film)
       *  2. Used in IIS sessions like /S(dfd346)/
      **/
      private static string URL_BALANCED_PARENS = @"\\($URL_VALID_GENERAL_PATH_CHARS+\\)";
      /** Valid end-of-path chracters (so /foo. does not gobble the period).
       *   2. Allow =&# for empty URL parameters and other URL-join artifacts
      **/
      private static string URL_VALID_PATH_ENDING_CHARS = @"[a-z0-9=_#/\\-\\+$LATIN_ACCENTS_CHARS]|(?:$URL_BALANCED_PARENS)";

      private static string URL_VALID_PATH = "(?:" +
        "(?:" +
          URL_VALID_GENERAL_PATH_CHARS + "*" +
          "(?:" + URL_BALANCED_PARENS + URL_VALID_GENERAL_PATH_CHARS + "*)*" +
          URL_VALID_PATH_ENDING_CHARS +
        ")|(?:@" + URL_VALID_GENERAL_PATH_CHARS + "+/)" +
      ")";

      private static string URL_VALID_URL_QUERY_CHARS = "[a-z0-9!?\\*'\\(\\);:&=\\+\\$/%#\\[\\]\\-_\\.,~\\|]";
      private static string URL_VALID_URL_QUERY_ENDING_CHARS = "[a-z0-9_&=#/]";
      private static string VALID_URL_PATTERN_STRING =
      "(" +                                                            //  $1 total match
        "(" + URL_VALID_PRECEEDING_CHARS + ")" +                       //  $2 Preceeding chracter
        "(" +                                                          //  $3 URL
          "(https?://)?" +                                             //  $4 Protocol (optional)
          "(" + URL_VALID_DOMAIN + ")" +                               //  $5 Domain(s)
          "(?::(" + URL_VALID_PORT_NUMBER +"))?" +                     //  $6 Port number (optional)
          "(/" + URL_VALID_PATH + "*)?" +                              //  $7 URL Path and anchor
          "(\\?" + URL_VALID_URL_QUERY_CHARS + "*" +                   //  $8 Query string
                  URL_VALID_URL_QUERY_ENDING_CHARS + ")?" +
        ")" +
      ")";

      private static string AT_SIGNS_CHARS = "@\\x{FF20}";

      private static string DOLLAR_SIGN_CHAR = "\\$";
      private static string CASHTAG = "[a-z]{1,6}(?:[._][a-z]{1,2})?";

      /* Begin public constants */

      public static Regex VALID_HASHTAG;
      public static int VALID_HASHTAG_GROUP_BEFORE = 1;
      public static int VALID_HASHTAG_GROUP_HASH = 2;
      public static int VALID_HASHTAG_GROUP_TAG = 3;
      public static Regex INVALID_HASHTAG_MATCH_END;

      public static Regex AT_SIGNS;
      public static Regex VALID_MENTION_OR_LIST;
      public static int VALID_MENTION_OR_LIST_GROUP_BEFORE = 1;
      public static int VALID_MENTION_OR_LIST_GROUP_AT = 2;
      public static int VALID_MENTION_OR_LIST_GROUP_USERNAME = 3;
      public static int VALID_MENTION_OR_LIST_GROUP_LIST = 4;

      public static Regex VALID_REPLY;
      public static int VALID_REPLY_GROUP_USERNAME = 1;

      public static Regex INVALID_MENTION_MATCH_END;

      public static Regex VALID_URL;
      public static int VALID_URL_GROUP_ALL          = 1;
      public static int VALID_URL_GROUP_BEFORE       = 2;
      public static int VALID_URL_GROUP_URL          = 3;
      public static int VALID_URL_GROUP_PROTOCOL     = 4;
      public static int VALID_URL_GROUP_DOMAIN       = 5;
      public static int VALID_URL_GROUP_PORT         = 6;
      public static int VALID_URL_GROUP_PATH         = 7;
      public static int VALID_URL_GROUP_QUERY_STRING = 8;

      public static Regex VALID_TCO_URL;
      public static Regex INVALID_URL_WITHOUT_PROTOCOL_MATCH_BEGIN;

      public static Regex VALID_CASHTAG;
      public static int VALID_CASHTAG_GROUP_BEFORE = 1;
      public static int VALID_CASHTAG_GROUP_DOLLAR = 2;
      public static int VALID_CASHTAG_GROUP_CASHTAG = 3;

      public TweetRegex () {
        try {
            VALID_HASHTAG = new Regex("(^|[^&" + HASHTAG_ALPHA_NUMERIC_CHARS + "])(#|\\x{FF03})(" + HASHTAG_ALPHA_NUMERIC + "*" + HASHTAG_ALPHA + HASHTAG_ALPHA_NUMERIC + "*)", RegexCompileFlags.CASELESS);
            INVALID_HASHTAG_MATCH_END = new Regex("^(?:[#＃]|://)");
            AT_SIGNS = new Regex("[" + AT_SIGNS_CHARS + "]");
            VALID_MENTION_OR_LIST = new Regex("([^a-z0-9_!#$%&*" + AT_SIGNS_CHARS + "]|^|RT:?)(" + AT_SIGNS.get_pattern() + "+)([a-z0-9_]{1,20})(/[a-z][a-z0-9_\\-]{0,24})?", RegexCompileFlags.CASELESS);
            VALID_REPLY = new Regex("^(?:[" + UNICODE_SPACES + "])*" + AT_SIGNS.get_pattern() + "([a-z0-9_]{1,20})", RegexCompileFlags.CASELESS);
            INVALID_MENTION_MATCH_END = new Regex("^(?:[" + AT_SIGNS_CHARS + LATIN_ACCENTS_CHARS + "]|://)");
            VALID_URL = new Regex(VALID_URL_PATTERN_STRING, RegexCompileFlags.CASELESS);
            VALID_TCO_URL = new Regex("^https?:\\/\\/t\\.co\\/[a-z0-9]+", RegexCompileFlags.CASELESS);
            INVALID_URL_WITHOUT_PROTOCOL_MATCH_BEGIN = new Regex("[-_./]$");
            VALID_CASHTAG = new Regex("(^|[" + UNICODE_SPACES + "])(" + DOLLAR_SIGN_CHAR + ")(" + CASHTAG + ")" +"(?=$|\\s|[" + PUNCT + "])", RegexCompileFlags.CASELESS);
        } catch (RegexError e) {
            stdout.printf ("Si llegamos a este punto, estamos jodidos.\n");
        }
      }
    }
}
